@@ -17,6 +17,7 @@ import (
 	"fmt"
 	"io"
 	"math"
+	"math/rand"
 	"strconv"
 	"testing"
 )
@@ -712,7 +713,7 @@ func TestShrink(t *testing.T) {
 	b.Set(80)
 	b.Shrink(70)
 	for _, word := range b.set {
-		if (word != 0) {
+		if word != 0 {
 			t.Error("word should be 0", word)
 		}
 	}
@@ -1153,60 +1154,53 @@ func TestComplement(t *testing.T) {
 }
 
 func TestIsSuperSet(t *testing.T) {
-	a := New(500)
-	b := New(300)
-	c := New(200)
+	test := func(name string, lenS, lenSS int, overrideS, overrideSS map[int]bool, want, wantStrict bool) {
+		t.Run(name, func(t *testing.T) {
+			s := New(uint(lenS))
+			ss := New(uint(lenSS))
 
-	// Setup bitsets
-	// a and b overlap
-	// only c is (strict) super set
-	for i := uint(0); i < 100; i++ {
-		a.Set(i)
-	}
-	for i := uint(50); i < 150; i++ {
-		b.Set(i)
-	}
-	for i := uint(0); i < 200; i++ {
-		c.Set(i)
+			l := lenS
+			if lenSS < lenS {
+				l = lenSS
+			}
+
+			r := rand.New(rand.NewSource(42))
+			for i := 0; i < l; i++ {
+				bit := r.Intn(2) == 1
+				s.SetTo(uint(i), bit)
+				ss.SetTo(uint(i), bit)
+			}
+
+			for i, v := range overrideS {
+				s.SetTo(uint(i), v)
+			}
+			for i, v := range overrideSS {
+				ss.SetTo(uint(i), v)
+			}
+
+			if got := ss.IsSuperSet(s); got != want {
+				t.Errorf("IsSuperSet() = %v, want %v", got, want)
+			}
+			if got := ss.IsStrictSuperSet(s); got != wantStrict {
+				t.Errorf("IsStrictSuperSet() = %v, want %v", got, wantStrict)
+			}
+		})
 	}
 
-	if a.IsSuperSet(b) {
-		t.Errorf("IsSuperSet fails")
-	}
-	if a.IsSuperSet(c) {
-		t.Errorf("IsSuperSet fails")
-	}
-	if b.IsSuperSet(a) {
-		t.Errorf("IsSuperSet fails")
-	}
-	if b.IsSuperSet(c) {
-		t.Errorf("IsSuperSet fails")
-	}
-	if !c.IsSuperSet(a) {
-		t.Errorf("IsSuperSet fails")
-	}
-	if !c.IsSuperSet(b) {
-		t.Errorf("IsSuperSet fails")
-	}
+	test("empty", 0, 0, nil, nil, true, false)
+	test("empty vs non-empty", 0, 100, nil, nil, true, false)
+	test("non-empty vs empty", 100, 0, nil, nil, true, false)
+	test("equal", 100, 100, nil, nil, true, false)
 
-	if a.IsStrictSuperSet(b) {
-		t.Errorf("IsStrictSuperSet fails")
-	}
-	if a.IsStrictSuperSet(c) {
-		t.Errorf("IsStrictSuperSet fails")
-	}
-	if b.IsStrictSuperSet(a) {
-		t.Errorf("IsStrictSuperSet fails")
-	}
-	if b.IsStrictSuperSet(c) {
-		t.Errorf("IsStrictSuperSet fails")
-	}
-	if !c.IsStrictSuperSet(a) {
-		t.Errorf("IsStrictSuperSet fails")
-	}
-	if !c.IsStrictSuperSet(b) {
-		t.Errorf("IsStrictSuperSet fails")
-	}
+	test("set is shorter, subset", 100, 200, map[int]bool{50: true}, map[int]bool{50: false}, false, false)
+	test("set is shorter, equal", 100, 200, nil, nil, true, false)
+	test("set is shorter, superset", 100, 200, map[int]bool{50: false}, map[int]bool{50: true}, true, true)
+	test("set is shorter, neither", 100, 200, map[int]bool{50: true}, map[int]bool{50: false, 150: true}, false, false)
+
+	test("set is longer, subset", 200, 100, map[int]bool{50: true}, map[int]bool{50: false}, false, false)
+	test("set is longer, equal", 200, 100, nil, nil, true, false)
+	test("set is longer, superset", 200, 100, nil, map[int]bool{150: true}, true, true)
+	test("set is longer, neither", 200, 100, map[int]bool{50: false, 150: true}, map[int]bool{50: true}, false, false)
 }
 
 func TestDumpAsBits(t *testing.T) {
@@ -1236,30 +1230,30 @@ func TestMarshalUnmarshalBinary(t *testing.T) {
 
 	aSetBit := uint(128)
 	a = New(256).Set(aSetBit)
-	aExpectedMarshaledSize := 8 /* length: uint64 */ + 4 * 8 /* set : [4]uint64 */
+	aExpectedMarshaledSize := 8 /* length: uint64 */ + 4*8 /* set : [4]uint64 */
 	aMarshaled, err := a.MarshalBinary()
 
 	if err != nil || aExpectedMarshaledSize != len(aMarshaled) || aExpectedMarshaledSize != a.BinaryStorageSize() {
-		t.Error("MarshalBinary failed to produce expected (", aExpectedMarshaledSize , ") number of bytes")
+		t.Error("MarshalBinary failed to produce expected (", aExpectedMarshaledSize, ") number of bytes")
 		return
 	}
 
 	shiftAmount := uint(72)
 	// https://github.com/bits-and-blooms/bitset/issues/114
-	for i := uint(0) ; i < shiftAmount; i++ {
+	for i := uint(0); i < shiftAmount; i++ {
 		a.DeleteAt(0)
 	}
 
-	aExpectedMarshaledSize = 8 /* length: uint64 */ + 3 * 8 /* set : [3]uint64 */
+	aExpectedMarshaledSize = 8 /* length: uint64 */ + 3*8 /* set : [3]uint64 */
 	aMarshaled, err = a.MarshalBinary()
 	if err != nil || aExpectedMarshaledSize != len(aMarshaled) || aExpectedMarshaledSize != a.BinaryStorageSize() {
-		t.Error("MarshalBinary failed to produce expected (", aExpectedMarshaledSize , ") number of bytes")
+		t.Error("MarshalBinary failed to produce expected (", aExpectedMarshaledSize, ") number of bytes")
 		return
 	}
 
 	copyBinary(t, a, b)
 
-	if b.Len() != 256 - shiftAmount || !b.Test(aSetBit - shiftAmount) {
+	if b.Len() != 256-shiftAmount || !b.Test(aSetBit-shiftAmount) {
 		t.Error("Shifted bitset is not copied correctly")
 	}
 }
@@ -1297,25 +1291,49 @@ func copyBinary(t *testing.T, from encoding.BinaryMarshaler, to encoding.BinaryU
 }
 
 func TestMarshalUnmarshalJSON(t *testing.T) {
-	a := New(1010).Set(10).Set(1001)
-	data, err := json.Marshal(a)
-	if err != nil {
-		t.Errorf(err.Error())
-		return
-	}
+	t.Run("value", func(t *testing.T) {
+		a := BitSet{}
+		a.Set(10).Set(1001)
+		data, err := json.Marshal(a)
+		if err != nil {
+			t.Errorf(err.Error())
+			return
+		}
 
-	b := new(BitSet)
-	err = json.Unmarshal(data, b)
-	if err != nil {
-		t.Errorf(err.Error())
-		return
-	}
+		b := new(BitSet)
+		err = json.Unmarshal(data, b)
+		if err != nil {
+			t.Errorf(err.Error())
+			return
+		}
 
-	// Bitsets must be equal after marshalling and unmarshalling
-	if !a.Equal(b) {
-		t.Error("Bitsets are not equal:\n\t", a.DumpAsBits(), "\n\t", b.DumpAsBits())
-		return
-	}
+		// Bitsets must be equal after marshalling and unmarshalling
+		if !a.Equal(b) {
+			t.Error("Bitsets are not equal:\n\t", a.DumpAsBits(), "\n\t", b.DumpAsBits())
+			return
+		}
+	})
+	t.Run("pointer", func(t *testing.T) {
+		a := New(1010).Set(10).Set(1001)
+		data, err := json.Marshal(a)
+		if err != nil {
+			t.Errorf(err.Error())
+			return
+		}
+
+		b := new(BitSet)
+		err = json.Unmarshal(data, b)
+		if err != nil {
+			t.Errorf(err.Error())
+			return
+		}
+
+		// Bitsets must be equal after marshalling and unmarshalling
+		if !a.Equal(b) {
+			t.Error("Bitsets are not equal:\n\t", a.DumpAsBits(), "\n\t", b.DumpAsBits())
+			return
+		}
+	})
 }
 
 func TestMarshalUnmarshalJSONWithTrailingData(t *testing.T) {
@@ -1479,6 +1497,43 @@ func TestClearAll(t *testing.T) {
 	}
 }
 
+func TestRankSelect(t *testing.T) {
+	u := []uint{2, 3, 5, 7, 11, 700, 1500}
+	b := BitSet{}
+	for _, v := range u {
+		b.Set(v)
+	}
+
+	if b.Rank(5) != 3 {
+		t.Error("Unexpected rank")
+		return
+	}
+	if b.Rank(6) != 3 {
+		t.Error("Unexpected rank")
+		return
+	}
+	if b.Rank(1500) != 7 {
+		t.Error("Unexpected rank")
+		return
+	}
+	if b.Select(0) != 2 {
+		t.Error("Unexpected select")
+		return
+	}
+	if b.Select(1) != 3 {
+		t.Error("Unexpected select")
+		return
+	}
+	if b.Select(2) != 5 {
+		t.Error("Unexpected select")
+		return
+	}
+
+	if b.Select(5) != 700 {
+		t.Error("Unexpected select")
+		return
+	}
+}
 func TestFlip(t *testing.T) {
 	b := new(BitSet)
 	c := b.Flip(11)
@@ -1892,3 +1947,19 @@ func TestReadFrom(t *testing.T) {
 		})
 	}
 }
+
+func TestSetAll(t *testing.T) {
+	test := func(name string, bs *BitSet, want uint) {
+		t.Run(name, func(t *testing.T) {
+			bs.SetAll()
+			if bs.Count() != want {
+				t.Errorf("expected %d bits to be set, got %d", want, bs.Count())
+			}
+		})
+	}
+
+	test("nil", nil, 0)
+	for _, length := range []uint{0, 1, 10, 63, 64, 65, 100, 640} {
+		test(fmt.Sprintf("length %d", length), New(length), length)
+	}
+}
